/* $Id: anton_absred.c,v 1.1.1.1 2010/05/02 13:52:33 uid523 Exp $ */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "xlts.h"
#include "messages.h"
#include "anton_absred.h"

static lts_t sig;

typedef struct trans {
	cell_t src;
	cell_t dest;
	uint32_t label;
} *trans_t;

void anton_abstime_reduce(lts_t lts){
	int mintime, currenttime;
	int first, found;
	int startpos, endpos;
	char *labelstring;
	int newstatenumber = 1;

	// Create a lookup array for reassigning numbers to states in the output LTS sig
	static int *table;
	// Create an OPEN set and a NEXT set to explore the LTS in a BFS manner
	static int *open, *next, *tmp;
	int open_current = 0;
	int next_current = 0;
	// Create a CLOSED set to prevent duplicates in OPEN/NEXT and duplicate work
	static int *closed;
	// Create a list of transitions for LTS sig
	//static trans_t **newlts;
	int *newlts_src;
	int *newlts_dest;
	uint32_t *newlts_label;
	
	int newlts_current = 0;
	
	sig=lts_create();

	lts_uniq(lts);
	lts_sort(lts);
	
	lts_set_type(lts,LTS_BLOCK);

	table=(int*)malloc(lts->states*sizeof(int));
	open=(int*)malloc(lts->states*sizeof(int));
	next=(int*)malloc(lts->states*sizeof(int));
	closed=(int*)malloc(lts->states*sizeof(int));
	//newlts=(trans_t**)malloc(lts->transitions*sizeof(trans_t*));
	newlts_src=(int*)malloc(lts->transitions*sizeof(int));
	newlts_dest=(int*)malloc(lts->transitions*sizeof(int));
	newlts_label=(uint32_t*)malloc(lts->transitions*sizeof(uint32_t));
	
	if ((next==NULL)||(table==NULL)||(open==NULL)||(newlts_src==NULL)||(newlts_dest==NULL)||(newlts_label==NULL)||(closed==NULL)) {
		Fatal(1,1,"out of memory in anton_abstime_reduce");
	}
	// Clear table, open and next
	int i,j,k;
	for(i=0;i<lts->states;i++) {
		table[i]=-1;
		open[i]=-1;
		next[i]=-1;
		closed[i]=0;
	}

	// Put the initial state in open, and start
	open[0] = lts->root;
	table[0] = 0;
	closed[0] = 1;
		
	while (open[open_current] != -1) {
		
		while (open[open_current] != -1) {
			mintime = 0;
			currenttime = 0;
			// Find the minimum time label of the state
			first = 1;
			// Range over the outgoing transitions looking for the minimum time stamp
		
			for(j=lts->begin[open[open_current]];j<lts->begin[open[open_current]+1];j++){
				labelstring = lts->label_string[lts->label[j]];
				// Search for end position of time stamp
				k = 0;
				found = 0;
				while (found == 0) {
					if (labelstring[k] == ')')
						found = 1;
					else if (labelstring[k] == '\0')
						found = 2;
					else {
						k++;
					}

				}
				if (found == 1) {
					endpos = k;
					// Search for starting position of time stamp
					k = 0;
					found = 0;
					while (found == 0) {
						if (labelstring[k] == '(')
							found = 1;
						else {
							k++;
						}
					}
					startpos = k+1;

					// Deal with the case that timing is 0, and there are more data pars
					// (i.e. do not count towards first ")", but first ",")
					k = 0;
					found = 0;
					while (found == 0) {
						if (labelstring[k] == ',')
							found = 1;
						else if (labelstring[k] == '\0')
							found = 2;
						else {
							k++;
						}

					}
					if (found = 1)
						if (k < endpos)
							endpos = k;
					currenttime = endpos - startpos;
					if (first) {
						mintime = currenttime;
						first = 0;
					}
					else if (currenttime < mintime)
						mintime = currenttime;
				}
			}
									
			// Prune away those transitions with a time label bigger than the minimum time
			for(j=lts->begin[open[open_current]];j<lts->begin[(open[open_current])+1];j++){
				labelstring = lts->label_string[lts->label[j]];
				// Search for end position of time stamp
				k = 0;
				found = 0;
				while (found == 0) {
					if (labelstring[k] == ')')
						found = 1;
					else if (labelstring[k] == '\0')
						found = 2;
					else {
						k++;
					}

				}
				if (found == 1) {
					endpos = k;
					// Search for starting position of time stamp
					k = 0;
					found = 0;
					while (found == 0) {
						if (labelstring[k] == '(')
							found = 1;
						else {
							k++;
						}

					}
					startpos = k+1;
					
					// Deal with the case that timing is 0, and there are more data pars
					// (i.e. do not count towards first ")", but first ",")
					k = 0;
					found = 0;
					while (found == 0) {
						if (labelstring[k] == ',')
							found = 1;
						else if (labelstring[k] == '\0')
							found = 2;
						else {
							k++;
						}

					}
					if (found = 1)
						if (k < endpos)
							endpos = k;
				
					// If the timing is minimum, we store the transition.
					if (mintime == endpos - startpos) {
						// Assign new state numbers if necessary.
						if (table[lts->dest[j]] == -1)
							table[lts->dest[j]] = newstatenumber++;
						newlts_src[newlts_current] = table[open[open_current]];
						newlts_dest[newlts_current] = table[lts->dest[j]];
						newlts_label[newlts_current] = lts->label[j];
						newlts_current++;
						// Put the dest state in the next set
						if (!closed[lts->dest[j]]) {
							next[next_current++] = lts->dest[j];
							closed[lts->dest[j]] = 1;
						}
					}
				}
				else {
					// Transition has no time stamp. Store it. 
					// Assign new state numbers if necessary.
					if (table[lts->dest[j]] == -1)
						table[lts->dest[j]] = newstatenumber++;
					newlts_src[newlts_current] = table[open[open_current]];
					newlts_dest[newlts_current] = table[lts->dest[j]];
					newlts_label[newlts_current] = lts->label[j];
					newlts_current++;
					// Put the dest state in the next set
					if (!closed[lts->dest[j]]) {
						next[next_current++] = lts->dest[j];
						closed[lts->dest[j]] = 1;
					}
				}
			}
			// Clean up the current position in the open set, and move to the next.
			open[open_current] = -1;
			open_current++;
		}
		// Switch open and next
		tmp = open;
		open = next;
		next = tmp;
		open_current = 0;
		next_current = 0;
	}
	
	// We are done. Write the transitions in newlts into sig.
	sig->label_string = lts->label_string;
	lts_free(lts);
	lts = lts_create();
	lts->label_string = sig->label_string;
	lts_set_size(lts,newstatenumber,newlts_current);
	lts_set_type(lts,LTS_LIST);
	for(i=0;i<newlts_current;i++){
		lts->src[i]=newlts_src[i];
		lts->label[i]=newlts_label[i];
		lts->dest[i]=newlts_dest[i];
	}
	
	free(table);
	free(newlts_src);
	free(newlts_dest);
	free(newlts_label);
	free(open);
	free(next);
	free(closed);
	lts_free(sig);
}
