/* $Id: Data.java,v 1.6 2003/04/17 14:12:44 bertl Exp $ */
package stateview;
import java.io.*;
import java.util.List;
import java.util.*;
import aterm.*;
import aterm.pure.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.table.*;
import java.awt.*;
import java.awt.event.*;
import att.grappa.*;
/** Representation of an summand in an LPO. */

class Summand {
          /** The printable condition. */
          public String condition;
          /** The printable action with arguments.  */
          public String action;
          /** The printable array of nextstates. */
          public String[] nextStates;
          /** The array is indexed by cluster_id. 
          * It is possible that more transitions belong to a summand.
          * Each transition belongs to a cluster. It is even possible that
          * more transitions belong to a (summand,cluster) pair. */
          public ArrayList[] transitions;
          /** The printable array of sumvariables. */
          public String[] sumVars;
          /** Constructor of a new Summand. */
          Summand(String condition, String action, String [] nextStates,
          ArrayList[] transitions, String[] sumVars) {
               // System.out.println("Condition:"+condition);
               this.condition = condition;
               this.action = action;
               this.nextStates = nextStates;
               this.transitions = transitions;
               this.sumVars = sumVars;
               for (int i=0;i<transitions.length;i++) 
                   transitions[i] = new ArrayList(2);
               }
               
/** prints the occurrence of a sum variable as bold.
@param term this will be checked 
@return the printable string 
*/ 
     
   String FormattedTerm(String term) {
       // System.out.println("term:"+term+":"+this.name);
       StringBuffer r = new StringBuffer("<HTML>");
       String delim = " ,()";
       StringTokenizer st = new StringTokenizer(term, delim, true);
       String s = st.nextToken();
           while (s!=null) {
                String nextString = null, t = "";
                while (st.hasMoreTokens()) {
                   String v = st.nextToken();
                   if (delim.indexOf(v)<0) {nextString=v;break;}
                   t = t.concat(v);
                   }
                int i;
                for (i=0;i<sumVars.length && !s.equals(this.sumVars[i]);i++);
                if (i<sumVars.length && (t.indexOf('(')==-1))
                   r.append("<I><U>"+ s + "</I></VAR>");
                else {
                  for (i=0;i<Data.parameters.size() && 
                    !s.equals(((Parameter) (Data.parameters.get(i))).name);i++);
                  if (i<Data.parameters.size()  && t.indexOf('(')==-1)
                   r.append("<I>"+ s + "</I>");
                  else
                     r.append(s);
                  }
           r.append(t);
           s = nextString;
           }
       r.append("</HTML>");
       return r.toString();
       }
          
          /** get printable transition
          * @param id cluster identifier
          * @return transitions in printable format
          */
               
          String getTransition(int id) {
               if (transitions[id]==null || transitions[id].size()==0) return "";
               Iterator it = transitions[id].iterator();
               StringBuffer buf = new StringBuffer(80);
               while (it.hasNext()) {
                   buf.append(((Transition) it.next()).from);
                   if (it.hasNext()) buf.append(",");
                   }
                   buf.append("->"+ ((Transition) transitions[id].get(0)).to);
               return buf.toString();
               }
          
          /** determines if a parameter occurs in this summand.
           * @param v the parameter which has to be checked.
           * @return  
           *         <ul>            
           *         <li> n, if the parameter occurs in condition
           *         <li> n+1, if the parameter occurs in action
           *         <li> p, if the position of the parameter is not equal to p 
           *                 and the parameter occurs in position p of nextstate
           *         <li> -1, if the parameter does not occur
           *         </ul>
           * in which n = number of parameters.
           */
           
          int Occurs(Parameter v) {
               if (v.OccursInTerm(condition)) return 
                      this.nextStates.length;
               if (v.OccursInTerm(action)) return
                      this.nextStates.length+1;
               for (int i=0;i<this.nextStates.length;i++)
               if (v.pos != i && v.OccursInTerm(nextStates[i])) return i;
               return -1;
               }
          } 

/** Representation of a summand in a stategraph. More transitions
can belong to a summand. For example, if a summand is member of more edges. 
A transition is member of exactly one edge. */ 
          
class Transition {
   String from, to;
   int sum;
   Transition(String f, int d, String t) {
      from = f; to = t; sum = d;
      }
   int intValue() {return sum;}
}

/** Representation of a process parameter. */
class Parameter {
/** The parameter name. */
   String name;
/** The names of the clusters separated by a blank in which this (parameter) 
*    occurs. */
   String occuringClusters;
/** The position of the parameter in the parameter list. */
   int pos;
/** Constructor of this class. */
   Parameter(String name, int pos) {
       this.name = name;
       this.pos = pos;
       }
/** checks the occurrence of this parameter in the argument.
@param term this will be checked for occurrence
@return true if parameter occurs, false otherwise 
*/      
    
    boolean OccursInTerm(String term) {
       // System.out.println("term:"+term+":"+this.name);
       String delim = " ,()";
       StringTokenizer st = new StringTokenizer(term, delim, true);
       String s = st.nextToken();
           while (s!=null) {
                String nextString = null, t = "";
                while (st.hasMoreTokens()) {
                   String v = st.nextToken();
                   if (delim.indexOf(v)<0) {nextString=v;break;}
                   t = t.concat(v);
                   }
                if (s.equals(this.name) && (t.indexOf('(')==-1))
                   return true;  
              t="";
              s = nextString;
              }
       return false;
       }
       
/** computes the clusters at which this parameter occurs.
@see #occuringClusters */       
   void occursCheck() {
      StringBuffer r = new StringBuffer();
      for (int i=0;i<Data.clusters.length;i++) {
        Cluster cluster = Data.clusters[i];
        Iterator it = cluster.smds.iterator();
        while (it.hasNext()) {
          int d = ((Integer) it.next()).intValue();
          Summand smd = Data.summands[d];
          if (smd.Occurs(this)>=0) {
              r.append(cluster.parameter.name+" ");
              break;
              }
          }
        }
      occuringClusters = 
          (r.length()>0?r.deleteCharAt(r.length()-1):r).toString();
          // System.out.println("ocuuringClusters:"+occuringClusters);
      }
}

/** List of parameters. */
class Parameters extends ArrayList {
   Parameters() {super();}
   Parameters(int size) {super(size);}
/** @param the parameter to be eliminated 
  * @return this parameter list minus the argument
  */ 
   Parameters Minus(Parameter par) {
       Parameters r = new Parameters(this.size());
       Iterator it = this.iterator();
       String name = par.name;
       while (it.hasNext()) {
           Parameter p = (Parameter) it.next();
           if (!p.name.equals(name)) r.add((Object) p);
           }
       return r;
       }
/** @param the list of parameters to be eliminated 
  * @return this parameter list minus the argument
  */        
    Parameters Minus(Parameters pars) {
       Parameters r = new Parameters(this.size());
       Iterator it1 = this.iterator();
       while (it1.hasNext()) {
           Parameter p1 = (Parameter) it1.next();
           String name = p1.name;
           Iterator it2 = pars.iterator();
           Parameter p2 = null;
           while (it2.hasNext()) {
              p2 = (Parameter) it2.next();
              if (p2.name.equals(name)) break;
              }
           if (p2!=null && !p2.name.equals(name)) {
               r.add((Object) p1);
               }
           }
       // System.out.println("r="+r);
       return r;
       }
/** @return the union from the argument and this list of parameters */
       
    Parameters Union(Parameters pars) {
       Parameters r = new Parameters(this.size());
       Iterator it1 = this.iterator();
       while (it1.hasNext()) {
           Parameter p1 = (Parameter) it1.next();
           String name = p1.name;
           Iterator it2 = pars.iterator();
           Parameter p2 = null;
           while (it2.hasNext()) {
              p2 = (Parameter) it2.next();
              if (p2.name.equals(name)) break;
              }
           if (p2==null || !p2.name.equals(name)) r.add((Object) p1);
           }
       it1 = pars.iterator();
       while (it1.hasNext()) {
           Parameter p2 = (Parameter) it1.next();
           r.add((Object) p2);
           }
       // System.out.println("r="+r);
       return r;
       }
 /** @return the union from the argument and this list of parameters */
       
    Parameters Intersection(Parameters pars) {
       Parameters r = new Parameters(this.size());
       Iterator it1 = this.iterator();
       while (it1.hasNext()) {
           Parameter p1 = (Parameter) it1.next();
           String name = p1.name;
           Iterator it2 = pars.iterator();
           Parameter p2 = null;
           while (it2.hasNext()) {
              p2 = (Parameter) it2.next();
              if (p2.name.equals(name)) {r.add((Object) p1);break;}
              }
           }
       return r;
       }
             
    void print() {
       System.out.println("parameters:");
       Iterator it1 = this.iterator();
          while (it1.hasNext()) {
             Parameter p1 = (Parameter) it1.next();
             System.out.print(" "+p1.name);
          }
    System.out.println("");     
    }     
   }
/** Set of transitions belonging to a cluster. */   
class Bundle {
    /** The transitions which are contained in this bundle. */
    Transition[] transitions;
    /** The edge which belong to this bundle. */
    private Edge line;
    /** The (action) label of the cluster (header symbol of action). */
    String label;
    /** The cluster of which this bundle is member. */
    public Cluster parent;
    /** Assigns an edge to this bundle. */
    public void setLine(Edge line) {
          this.line = line;
          Data.lineBundle.put((Object) line, this);
          }
    /** @return the edge from this bundle. */
    public Edge getLine() {return this.line;}
    /**  @param  the parameter which will be tested for occurrence
     *    @return true if argument occurs in a condition of one of 
     *    the summands */ 
    public boolean occursInCondition(Parameter par) {
          for (int j = 0;j<transitions.length;j++) {
            Transition transition = transitions[j];
            Summand summand = Data.summands[transition.sum];
            if (par.OccursInTerm(summand.condition)) return true;
            }
          return false;
          }
          
/**  @param  the parameter which will be tested for occurrence
*    @return true if argument occurs in a action of one of 
*    the summands */
      
    public boolean occursInAction(Parameter par) {
          for (int j = 0;j<transitions.length;j++) {
            Transition transition = transitions[j];
            Summand summand = Data.summands[transition.sum];
            if (par.OccursInTerm(summand.action)) return true;
            }
          return false;
          }
          
    /**  @param  the parameter which will be tested for occurrence
     *    @return true if argument occurs in a next state component of one of 
     *    the summands */
     public boolean occursInNextState(Parameter par) {
          for (int j = 0;j<transitions.length;j++) {
            Transition transition = transitions[j];
            Summand summand = Data.summands[transition.sum];    
            for (int s=0;s<summand.nextStates.length;s++) {       
               if (par.OccursInTerm(summand.nextStates[s]) && (s!=par.pos
               /* Modification adviced by Jaco vd Pol */ 
               || s==par.pos && !par.name.equals(summand.nextStates[s]))) 
               return true;        
            }
          }
       return false;
       }
/** @param par the parameter which will be tested if it is overwritten
*    in all summands */     
     public boolean isWritten(Parameter par) {
       // System.err.println("IsWritten "+par.name);
       for (int j = 0;j<transitions.length;j++) {
            Transition transition = transitions[j];
            Summand summand = Data.summands[transition.sum];
            // System.err.println(summand.nextStates[par.pos]);
            //if (par.OccursInTerm(summand.nextStates[par.pos])) 
            //     return false;
            /* Modification adviced by Jaco vd Pol */ 
            if (!par.name.equals(summand.nextStates[par.pos])) return true;
            }
       return false;
       }
/** computes and sets the color of the edges, dictated by the occurrence
*   of the selected parameters in the summands of the edge */ 
void setColor() { 
       Parameter par = this.parent.selectedPar;
       if (par == null) return;     
       Edge line = this.getLine();
       line.setAttribute(GrappaLine.COLOR_ATTR,  
         (Object) (this.isWritten(par)?Color.blue:Color.black)); 
       if (this.occursInAction(par) || this.occursInNextState(par)) 
         line.setAttribute(GrappaLine.FONTCOLOR_ATTR, (Object) (Color.red));
       if (this.occursInCondition(par)) 
         line.getTail().setAttribute(GrappaLine.COLOR_ATTR, 
           (Object) (Color.red));
       }
       
/** computes and clears the color of the edges, dictated by the occurrence
*   of the selected parameters in the summands of the edge */ 
void clearColor() { 
       Parameter par = this.parent.selectedPar;
       if (par == null) return;
       Edge line = this.getLine();
       line.setAttribute(GrappaLine.COLOR_ATTR, Color.black);
       line.getTail().setAttribute(GrappaLine.COLOR_ATTR, 
              (Object) Color.black);
       line.setAttribute(GrappaLine.FONTCOLOR_ATTR, 
              (Object) Color.black);
       // line.setAttribute(GrappaLine.STYLE_ATTR, "solid");
       }   
                  
 /** Class constructor belonging to an Aterm representation of an edge. */   
    Bundle(ATerm edge) {
       // System.err.println("QQQ edge = "+edge);
       ATermList ls = (ATermList) (((ATermAppl) edge).getArgument(1));
       ATerm from = ((ATermList) (((ATermAppl) edge).getArgument(0))).getFirst();
       ATerm to = ((ATermList) ((ATermAppl) edge).getArgument(2)).getFirst();
       transitions = new Transition[ls.getLength()];
       if (ls.getLength()>0) {
           int summand = ((ATermInt) ls.getFirst()).getInt();
           String action = Data.summands[summand].action;
           int index = action.indexOf('(');
           label = index<0?action:action.substring(0, index);
           }
       for (int i=0;!ls.isEmpty();ls = ls.getNext(),i++) {
          transitions[i]= new Transition(
               Data.print(from).toString(),
               ((ATermInt) ls.getFirst()).getInt(),
               Data.print(to).toString()
          );
          }
       }
     
/** returns the number of transitons */
   
    int size() {return transitions.length;}
    }

/** A subset of summands defined by an process (state) parameter. */
class Cluster {
/** The defining parameter which contains the name of the cluster, which 
 *  is called here cluster parameter. */
    Parameter parameter;
/** The index of the cluster. */
    int id;
/** The edges of the cluster. */
    Bundle[] edges;
/** Mapping from state value to Grappa Node. */
    Hashtable table;
/** The set of summands which belong to this cluster.  */
    TreeSet smds;
    final Integer[] smdsA;
/** The complement of smds. */
    TreeSet smdsC;
/** The parameters which occurs in the summands of this cluster and 
*    are no cluster parameters. */
    Parameters pars;
/** The parameters which does not occur in the summands of the compement
    this cluster. */
    Parameters purePars;
/** This parameters may occur in complement of this cluster. */
    Parameters dirtyPars;
/** The selected parameter */
    Parameter selectedPar;
/** The list of neighbouring clusters. */
    ArrayList neighbours;
/** The graph of this cluster. */
    Graph graph;
    StateGraph stategraph;
/** Class constructor of a cluster.
@param parameter defines this cluster
*/   
    Cluster(Parameter parameter,  Bundle[] edges, int id,
    Graph graph) {
     this.edges = edges;
     this.parameter = parameter;
     this.graph = graph;
     for (int i=0;i<edges.length;i++)
         edges[i].parent = this;
     this.table = new Hashtable();
     this.smds =  new TreeSet();
     this.neighbours = new ArrayList(Data.clusters.length);
     this.id = id;
     for (int i=0;i<edges.length;i++) {
       Bundle edge = edges[i];
       for (int j=0;j<edge.transitions.length;j++) {
           smds.add(new Integer(edge.transitions[j].sum));
           // System.out.println("seq1="+edge.transitions[j].sum);
           }
     }
    smdsA= (Integer[]) smds.toArray(new Integer[0]);
    smdsC = Complement(smds);
    // Data.parameters.print();
    // Data.clusterNames.print(); 
    Parameters pars = Data.parameters.Minus(Data.clusterNames);
    this.pars =  Occurs(pars);
    this.purePars = pars.Intersection(
         this.NotOccursInComplement(this.pars));
    this.dirtyPars = this.pars.Minus(this.purePars);
    // this.dirtyPars.print();
    this.pars= this.purePars.Union(this.dirtyPars);
    }
    
/** adds neighbours to this cluster */
    
    void addNeighbours() {
       for (int i=0;i<Data.clusters.length;i++)
       if (i != this.id) {
          if (Data.clusters[i].Occurs(this.parameter))
          neighbours.add(Data.clusters[i]);
          }
       }

/** @return a subset of the argument which occur in the summands of this cluster */       
    public Parameters Occurs(Parameters pars) {
      // pars.print();
      Parameters r = new Parameters(Data.npars);
      // System.out.println("size "+ smds.size());
      Iterator it1 = pars.iterator();
      while (it1.hasNext()) {
        Parameter par = (Parameter) it1.next();
        // Bert if (Occurs(par)  && NotOccursInComplement(par)) r.add(par);
        if (Occurs(par)) r.add(par);
        }
      // r.print();
      return r;
      }
      
/** @return a subset of the argument which not occurs in the summands of the 
compement of this  cluster */
       
    public Parameters NotOccursInComplement(Parameters pars) {
      // pars.print();
      Parameters r = new Parameters(Data.npars);
      // System.out.println("size "+ smds.size());
      Iterator it1 = pars.iterator();
      while (it1.hasNext()) {
        Parameter par = (Parameter) it1.next();
        // Bert if (Occurs(par)  && NotOccursInComplement(par)) r.add(par);
        if (NotOccursInComplement(par)) r.add(par);
        }
      // r.print();
      return r;
      }
      
/** computes the complement of the set of summands of this cluster */
    
     private TreeSet Complement(TreeSet s) {
         TreeSet r = new TreeSet();
         for (int i=0;i<Data.nrows;i++)
            r.add(new Integer(i));
         Iterator it = s.iterator();
         while (it.hasNext()) 
           r.remove(it.next());
         return r;
         }
/** tests if a parameter occurs in one of the summands of this cluster. 
*   @param par parameter which is subject of the test
*   @return true if parameter occurs, false otherwise
*/         
     public boolean Occurs(Parameter par) {
        Iterator it2 = smds.iterator();
        while (it2.hasNext()) {
            int d = ((Integer) it2.next()).intValue();
            // System.out.println("seq="+d);
            Summand smd = Data.summands[d];
            if (smd.Occurs(par)>=0) {
              return true;
              }
            }
        return false;
        }
        
/** tests if a parameter not occurs in one of the complement summands of 
*   this cluster. 
*   @param par parameter which is subject of the test
*   @return true if parameter not occurs, false otherwise
*/        
      public boolean NotOccursInComplement(Parameter par) {
        Iterator it2 = smdsC.iterator();
        while (it2.hasNext()) {
            int d = ((Integer) it2.next()).intValue();
            // System.out.println("seq="+d);
            Summand smd = Data.summands[d];
            if (smd.Occurs(par)>=0) {
              return false;
              }
            }
        return true;
        }
}

/** The static class which contains the basic data. */            
public class Data {
/** The result of the ATerm initialisation. */
  public static ATermFactory f;
/** The summands of the LPO. */
  public static Summand summands[];
/** The number of summands. */
  public static int nrows;
/** The number of process parameters. */
  public static int npars;
/** The process parameters. */
  public static ATermList pars;
/** The array of the clusters. */
  public static Cluster clusters[];
/** The cluster diagram. */
  public static Graph clustergraph;
 
/** The arraylist of defining parameters of the clusters. */  
  public static Parameters clusterNames;
/** The arraylist of all parameters. */  
  public static Parameters parameters;
/** Mapping from Edge to Bundle. */
  public static Hashtable lineBundle = new Hashtable(200);
/** 
*    parses dot output file.
*    @param rd graph defined in dot format
*    @return grappa graph
*/       
   private static Graph parse(StringReader rd) {
      Graph graph = null;
      Parser program = new Parser(rd, new PrintWriter(System.err, true));
      try {
	program.parse();
      } catch(Exception ex) {
	  System.err.println("Exception: " + ex.getMessage());
	  ex.printStackTrace(System.err);
	  System.exit(1);
      }
    graph = program.getGraph();
    // graph.printGraph(new PrintWriter(System.out));
    // graph.printGraph(new PrintWriter(System.out));
    System.err.println("The graph contains " + 
    graph.countOfElements(Grappa.NODE|Grappa.EDGE|Grappa.SUBGRAPH) 
     + " elements.");
    //graph.setMenuable(true);
    // graph.setErrorWriter(new PrintWriter(System.err,true));
    return graph;     
    }
    
  public static ATerm print(ATerm t) {
     String s = ((ATermAppl) t).getAFun().getName();
     int n = ((ATermAppl) t).getAFun().getArity(), i;
     s = s.substring(0, s.indexOf('#'));
     // System.out.println(s);
     ATerm arg[] = new ATerm[n];
     for (i=0;i<n;i++) 
          arg[i] = print(((ATermAppl) t).getArgument(i));
     return (ATerm) f.makeAppl(f.makeAFun(s, n, false),  arg);
     }
       
  public static Parameters parNames() {
      Parameters r = new Parameters(pars.getLength()); 
        int i = 0;
        for (ATermList pars = Data.pars;!pars.isEmpty();pars=pars.getNext(), i++) {
           ATerm par = Data.print(((ATermAppl) pars.getFirst()).getArgument(0));
           String header = ((ATermAppl)par).getAFun().getName();
           r.add(new Parameter(header, i));
           }
      return r;
      } 
     
  public static ATermList printlist(ATermList ts) {
     ATermList r = f.makeList();
     for (;!ts.isEmpty();ts=ts.getNext())
        r = r.insert(print(ts.getFirst()));
     return r.reverse();
  }
  
  public static ATerm printaction(ATerm name, ATermList args) {
     AFun s = f.makeAFun(((ATermAppl) name).getAFun().getName(), args.getLength(), false);
     return (ATerm) f.makeAppl(s, printlist(args));
     }
     
  public static ATermList parameterlist(ATermList ts) {
     ATermList r = f.makeList();
     for (;!ts.isEmpty();ts=ts.getNext()) {
     ATerm par = pars.elementAt(((ATermInt) ts.getFirst()).getInt());
        r = r.insert(Data.print(((ATermAppl) par).getArgument(0)));
        }
     return r.reverse();
     }
  
  public static String DotFile(ATerm pair) {
     return ((ATermAppl) ((ATermAppl) pair).getArgument(0)).
             getAFun().getName();
     }
     
  public static Bundle[] getEdges(ATermList es) {
     Bundle[] bundle = new Bundle[es.getLength()];
     for (int i=0;!es.isEmpty();es = es.getNext(),i++) 
         bundle[i] = new Bundle(es.getFirst());
     return bundle;     
     }
        
 /** adds transitions belonging to a cluster to the summands of 
     that cluster. Summands can belong to more clusters. */

  static private void updateSummands(Cluster cluster) {
       for (int i=0;i<cluster.edges.length;i++) {
            Bundle edge = cluster.edges[i];
            for (int j=0;j<edge.transitions.length;j++)
               Data.summands[edge.transitions[j].sum].transitions[cluster.id]
               .add((Object) edge.transitions[j]);
            }
       }
         
static void MakeNextStates(ATermAppl spec) {
        ATermAppl proc = (ATermAppl) spec.getArgument(1);
        ATermList sums = (ATermList) proc.getArgument(2);
        pars = (ATermList) proc.getArgument(1);
        // System.out.println("Clusters"+clusters);
        nrows =  sums.getLength();
        npars = pars.getLength();
        summands = new Summand[nrows];
        for (int i=0;!sums.isEmpty();sums=sums.getNext(),i++) {
            String[] nextStates = new String[npars];
            ATermList args = (ATermList) 
            ((ATermAppl) ((ATermAppl) sums.getFirst()).getArgument(3)).getArgument(0);
            ATermList par = pars;
            String action = Data.printaction(((ATermAppl) sums.getFirst()).getArgument(1),
            (ATermList) ((ATermAppl) sums.getFirst()).getArgument(2)).toString();
            String condition = Data.print(((ATermAppl) sums.getFirst()).getArgument(4)).toString();
            ATermList sumvars = (ATermList) ((ATermAppl) sums.getFirst()).getArgument(0);
            String[] sumVars = new String[sumvars.getLength()];
            for (int j=0;!sumvars.isEmpty();sumvars=sumvars.getNext(),j++) 
                 {sumVars[j] = Data.print(((ATermAppl) sumvars.getFirst()).getArgument(0)).toString();
                 } 
            for (int j=0;!args.isEmpty();args=args.getNext(), par=par.getNext(),j++) {
              nextStates[j] = print(args.getFirst()).toString();
              }
            summands[i] = new Summand(condition, action, nextStates,   
                new ArrayList[clusters.length],sumVars);         
            }
        }
            
  public static void makeData(InputStream input) {
        f = new aterm.pure.PureFactory();
        try {
        // System.out.println("readFromFile");
        ATermList in =  (ATermList) (f.readFromFile(input));
        // System.out.println("redFromFile"); 
        clustergraph = 
        parse(new StringReader(DotFile(in.elementAt(0))));
        ATermList graphlist =  (ATermList) in.elementAt(1);
        clusters = new Cluster[graphlist.getLength()];
        Graph[] graphs = new Graph[clusters.length];
        Bundle[][] edges = new Bundle[graphs.length][];
        MakeNextStates((ATermAppl)in.elementAt(2));
        // System.out.println("Test0 "+ graphs.getLength());
        Data.parameters =  Data.parNames();
        Data.clusterNames = new Parameters(clusters.length);
        for (int i=0;!graphlist.isEmpty();graphlist=graphlist.getNext(),i++) {
            ATermList statepars = 
            (ATermList) ((ATermAppl) ((ATermAppl)graphlist.getFirst()).
            getArgument(1)).getArgument(0);
            ATermList es = 
            (ATermList) ((ATermAppl) ((ATermAppl)graphlist.getFirst()).
            getArgument(1)).getArgument(2);
            int seq = ((ATermInt) statepars.getFirst()).getInt();
            Object parameter = Data.parameters.get(seq);
            Data.clusterNames.add(parameter);
            graphs[i] = parse(new StringReader(DotFile(graphlist.getFirst())));
            edges[i] = getEdges(es);
            }    
        for (int i=0;i<graphs.length;i++) {
            Parameter parameter = (Parameter) Data.clusterNames.get(i);
            clusters[i] = new Cluster(parameter,edges[i],i, graphs[i]);
            Data.updateSummands(clusters[i]);
            }
        Iterator it = Data.parameters.iterator();
        while (it.hasNext()) {
              ((Parameter) it.next()).occursCheck();
              }
        for (int i=0;i<Data.clusters.length;i++) 
            Data.clusters[i].addNeighbours();
        } catch(IOException fnf) {
	    System.err.println(fnf.toString());
	    System.exit(1);
	} 
     }
 }
 
