/* $Id: StateGraph.java,v 1.4 2003/04/16 11:47:00 bertl Exp $ */
package stateview;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.*;
import java.lang.String;
import att.grappa.*;
/** The class which provides nice list elements. */
class MyCellRenderer extends JLabel implements ListCellRenderer {
     private Border border;
     private Cluster cluster;
     public MyCellRenderer(Cluster cluster) {
         setOpaque(true);
         this.cluster = cluster;
         border = BorderFactory.createBevelBorder(BevelBorder.RAISED);
     }
     public Component getListCellRendererComponent(
         JList list,
         Object value,
         int index,
         boolean isSelected,
         boolean cellHasFocus)
     {
     // System.out.println("aap:"+index+" "+cluster.purePars.size()+" "+
     // cluster.dirtyPars.size());
         setText(value.toString());
         this.setBorder(border);
         setBackground(isSelected ? Color.blue : 
         (index-1<cluster.purePars.size()?Color.white:Color.yellow));
         setForeground(isSelected ? Color.white : Color.black);
         return this;
     }
 }
/**
* The class defining a stategraph belonging to a cluster.
* @author Bert Lisser
*/        
class StateGraph extends JFrame implements ListSelectionListener {
        /**
        * The grappa panel of the stategraph.
        */
	GrappaPanel gp;
        /**
        * The graph of the stategraph.
        */
	Graph graph = null;
        /**
        * The cluster to which the state graph belongs.
        * 
        */
        Cluster cluster;
        /** The reactions on the graph */
        final React react;
        /**
        * The array of names variables which occur only in this 
        * stategraph. This array is entry at the call of JList.
        */
        String[] names;
        /** The scrollpane which contains the JList. */
        JScrollPane scrollpane;
        /** The border around the JList. */
        private Border border = 
                BorderFactory.createBevelBorder(BevelBorder.RAISED);
        /** @return the belonging grappa graph */
        public Graph getGraph() {return graph;}
        /**
        * Class constructor.
        * @param cluster the cluster to which the state graph belongs
        */
	public StateGraph(Cluster cluster) {
            super(cluster.parameter.name);
            scrollpane = new JScrollPane();
            this.cluster = cluster;
            graph = cluster.graph;
            /* GrappaSupport.setHighlight(e.nextGraphElement(), 1, 
               GrappaConstants.HIGHLIGHT_ON); */
            Dimension dim = new Dimension(700,700);
	    gp = new GrappaPanel(graph) {
               public Dimension getMaximumSize() {return getPreferredSize();}
            };
            this.react = new React(cluster);
            gp.addGrappaListener(react);
            
            Stateview.ComputeBounds(gp, graph);
            Rectangle bound = new Rectangle(dim),
            port = new Rectangle(new Dimension(gp.getPreferredSize()));
            JViewport vp = new JViewport() {
                 public Dimension getMaximumSize() {return getPreferredSize();}
            };
            vp.setView(gp);
            bound = port.intersection(bound);
            if (bound.equals(port))  
               gp.setScaleToFit(true);   
            else {
                if (port.getSize().width>1000 || 
                    port.getSize().height>1000) {
                  gp.setScaleToFit(false); 
                  
                  }
                else gp.setScaleToFit(true);
                vp.setPreferredSize(bound.getSize());
                }
            scrollpane.setViewport(vp);
            // scrollpane.setDoubleBuffered(true);
            getContentPane().add("Center", scrollpane);
 //          
 // Add List of parameters which defines the cluster
 //
            names = new String[cluster.pars.size()+1];
            names[0] = "RESET";
            int i = 1;
            Iterator it = cluster.pars.iterator();
            while (it.hasNext()) names[i++] = ((Parameter) it.next()).name;
            JList list = new JList(names) {
            public Dimension getMaximumSize() {return getPreferredSize();}
            };
            JScrollPane pane = new JScrollPane(list) {
               public Dimension getMaximumSize() {return getPreferredSize();}
            };
            list.setCellRenderer(new MyCellRenderer(cluster));
            list.setBorder(border);
            list.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
            list.addListSelectionListener(this);
            pane.getViewport().setPreferredSize(list.getMaximumSize());
            Box box = Box.createVerticalBox();
            //panel.setLayout(new BorderLayout());
            //panel.add("Center",new JLabel(""));
            Component c = Box.createVerticalGlue();
            box.add(c);
            box.add(pane);
            c = Box.createVerticalGlue();
            box.add(c);
            getContentPane().add("East", box);
            getContentPane().addMouseListener(new MouseAdapter() {
		   public void mouseEntered(MouseEvent e) {
                        // System.out.println("aap:"+e);
			react.clearHighlight();
		    }
		});
            this.setJMenuBar(new MyJMenuBar(cluster));    
            pack();
            addWindowListener(new WindowAdapter() {
		    public void windowClosing(WindowEvent wev) {
			Window w = wev.getWindow();
			w.setVisible(false);
			if (react.dialog!=null) react.dialog.setVisible(false);
		    }
		});	
            // list.setSelectedIndex(0);
            // cluster.stategraph.repaint(); 	    	    
     }
     
     private void printAttributes(Element line) {
           Enumeration e = line.getAttributePairs();
           while (e.hasMoreElements()) System.out.println("QQ:"+e.nextElement());
           }
           
/** defines the reactions at clicking on items of the list. 
*    Edges are colored:
*   <ul>
*   <li> red, if variable occurs in  one summand belonging to that edge
*   <li> blue, if variable is overwritten in all summands belonging to that edge
*   </ul>
*/          
     public void valueChanged(ListSelectionEvent e) {
      if (!e.getValueIsAdjusting()) {
         JList list = (JList) (e.getSource());
         Bundle[] edges= this.cluster.edges;
         if (edges.length==0) return;
         // printAttributes(edges[0].getLine());
         for (int i=0;i<edges.length;i++) {
            edges[i].clearColor();
            }
         int pos = list.getAnchorSelectionIndex()-1;
         if (pos>=0) {
         this.cluster.selectedPar = (Parameter) (this.cluster.pars.get(pos));
         Parameter par = this.cluster.selectedPar;
           for (int i=0;i<edges.length;i++) {
             Bundle edge = edges[i];
             // System.out.println("QQ:"+edge.label);
             edge.setColor();     
             }
            }
         else {
           Stateview.thread.clear();
           this.cluster.selectedPar = null;
           }
       cluster.stategraph.repaint(); 
       }
    }
}     
